@@ -17,13 +17,16 @@ from __future__ import absolute_import
 import itertools
 import operator
 import os.path
+import uuid
 
+import oslo_db.api
 from oslo_db import exception
 from oslo_db.sqlalchemy import models
 from oslo_db.sqlalchemy import session
 from oslo_db.sqlalchemy import utils as oslo_db_utils
 import six
 import sqlalchemy
+import sqlalchemy_utils
 from stevedore import extension
 
 from gnocchi import exceptions
@@ -46,14 +49,15 @@ def get_resource_mappers(ext):
         resource_ext = ext.plugin
         resource_history_ext = ResourceHistory
     else:
+        tablename = getattr(ext.plugin, '__tablename__', ext.name)
         resource_ext = type(str(ext.name),
                             (ext.plugin, base.ResourceExtMixin, Resource),
-                            {"__tablename__": ext.name})
+                            {"__tablename__": tablename})
         resource_history_ext = type(str("%s_history" % ext.name),
                                     (ext.plugin, base.ResourceHistoryExtMixin,
                                      ResourceHistory),
                                     {"__tablename__": (
-                                        "%s_history" % ext.name)})
+                                        "%s_history" % tablename)})
 
     return {'resource': resource_ext,
             'history': resource_history_ext}
@@ -125,18 +129,20 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
             if session.query(ArchivePolicy).filter(
                     ArchivePolicy.name == name).delete() == 0:
                 raise indexer.NoSuchArchivePolicy(name)
-        except exception.DBError as e:
-            # TODO(jd) Add an exception in oslo.db to match foreign key
-            # violations
-            if isinstance(e.inner_exception, sqlalchemy.exc.IntegrityError):
+        except exception.DBReferenceError as e:
+            if (e.constraint ==
+               'fk_metric_archive_policy_name_archive_policy_name'):
                 raise indexer.ArchivePolicyInUse(name)
+            raise
 
-    def get_metrics(self, uuids):
+    def get_metrics(self, uuids, active_only=True):
         if not uuids:
             return []
         session = self.engine_facade.get_session()
         query = session.query(Metric).filter(Metric.id.in_(uuids)).options(
             sqlalchemy.orm.joinedload('resource'))
+        if active_only:
+            query = query.filter(Metric.status == 'active')
 
         metrics = list(query.all())
         session.expunge_all()
@@ -160,7 +166,8 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
 
     def list_archive_policy_rules(self):
         session = self.engine_facade.get_session()
-        aps = session.query(ArchivePolicyRule).all()
+        aps = session.query(ArchivePolicyRule).order_by(
+            ArchivePolicyRule.metric_pattern.desc()).all()
         session.expunge_all()
         return aps
 
@@ -172,14 +179,9 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
 
     def delete_archive_policy_rule(self, name):
         session = self.engine_facade.get_session()
-        try:
-            if session.query(ArchivePolicyRule).filter(
-                    ArchivePolicyRule.name == name).delete() == 0:
-                raise indexer.NoSuchArchivePolicyRule(name)
-        except exception.DBError as e:
-            # TODO(prad): fix foreign key violations when oslo.db supports it
-            if isinstance(e.inner_exception, sqlalchemy.exc.IntegrityError):
-                raise indexer.ArchivePolicyRuleInUse(name)
+        if session.query(ArchivePolicyRule).filter(
+                ArchivePolicyRule.name == name).delete() == 0:
+            raise indexer.NoSuchArchivePolicyRule(name)
 
     def create_archive_policy_rule(self, name, metric_pattern,
                                    archive_policy_name):
@@ -199,8 +201,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
 
     def create_metric(self, id, created_by_user_id, created_by_project_id,
                       archive_policy_name,
-                      name=None, resource_id=None,
-                      details=False):
+                      name=None, resource_id=None):
         m = Metric(id=id,
                    created_by_user_id=created_by_user_id,
                    created_by_project_id=created_by_project_id,
@@ -209,17 +210,20 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
                    resource_id=resource_id)
         session = self.engine_facade.get_session()
         session.add(m)
-        session.flush()
-        if details:
-            # Fetch archive policy
-            m.archive_policy
+        try:
+            session.flush()
+        except exception.DBReferenceError as e:
+            if (e.constraint ==
+               'fk_metric_archive_policy_name_archive_policy_name'):
+                raise indexer.NoSuchArchivePolicy(archive_policy_name)
+            raise
         session.expunge_all()
         return m
 
     def list_metrics(self, user_id=None, project_id=None, details=False,
-                     **kwargs):
+                     status='active', **kwargs):
         session = self.engine_facade.get_session()
-        q = session.query(Metric)
+        q = session.query(Metric).filter(Metric.status == status)
         if user_id is not None:
             q = q.filter(Metric.created_by_user_id == user_id)
         if project_id is not None:
@@ -273,13 +277,12 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
         session.expunge_all()
         return r
 
+    @oslo_db.api.retry_on_deadlock
     def update_resource(self, resource_type,
                         resource_id, ended_at=_marker, metrics=_marker,
                         append_metrics=False,
+                        create_revision=True,
                         **kwargs):
-
-        now = utils.utcnow()
-
         resource_cls = self._resource_type_to_class(resource_type)
         resource_history_cls = self._resource_type_to_class(resource_type,
                                                             "history")
@@ -297,12 +300,15 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
                 if r is None:
                     raise indexer.NoSuchResource(resource_id)
 
-                # Build history
-                rh = resource_history_cls()
-                for col in sqlalchemy.inspect(resource_cls).columns:
-                    setattr(rh, col.name, getattr(r, col.name))
-                rh.revision_end = now
-                session.add(rh)
+                if create_revision:
+                    # Build history
+                    rh = resource_history_cls()
+                    for col in sqlalchemy.inspect(resource_cls).columns:
+                        setattr(rh, col.name, getattr(r, col.name))
+                    now = utils.utcnow()
+                    rh.revision_end = now
+                    session.add(rh)
+                    r.revision_start = now
 
                 # Update the resource
                 if ended_at is not _marker:
@@ -315,8 +321,6 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
                                     resource_type, "ended_at", ended_at)
                     r.ended_at = ended_at
 
-                r.revision_start = now
-
                 if kwargs:
                     for attribute, value in six.iteritems(kwargs):
                         if hasattr(r, attribute):
@@ -328,7 +332,8 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
                 if metrics is not _marker:
                     if not append_metrics:
                         session.query(Metric).filter(
-                            Metric.resource_id == resource_id).update(
+                            Metric.resource_id == resource_id,
+                            Metric.status == 'active').update(
                                 {"resource_id": None})
                     self._set_metrics_for_resource(session, r, metrics)
         except exception.DBConstraintError as e:
@@ -345,31 +350,55 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
 
     @staticmethod
     def _set_metrics_for_resource(session, r, metrics):
-        for name, metric_id in six.iteritems(metrics):
-            try:
-                update = session.query(Metric).filter(
-                    Metric.id == metric_id,
-                    Metric.created_by_user_id == r.created_by_user_id,
-                    Metric.created_by_project_id == r.created_by_project_id,
-                ).update({"resource_id": r.id, "name": name})
-            except exception.DBDuplicateEntry:
-                raise indexer.NamedMetricAlreadyExists(name)
-            if update == 0:
-                raise indexer.NoSuchMetric(metric_id)
+        for name, value in six.iteritems(metrics):
+            if isinstance(value, uuid.UUID):
+                try:
+                    update = session.query(Metric).filter(
+                        Metric.id == value,
+                        Metric.status == 'active',
+                        (Metric.created_by_user_id
+                         == r.created_by_user_id),
+                        (Metric.created_by_project_id
+                         == r.created_by_project_id),
+                    ).update({"resource_id": r.id, "name": name})
+                except exception.DBDuplicateEntry:
+                    raise indexer.NamedMetricAlreadyExists(name)
+                if update == 0:
+                    raise indexer.NoSuchMetric(value)
+            else:
+                ap_name = value['archive_policy_name']
+                m = Metric(id=uuid.uuid4(),
+                           created_by_user_id=r.created_by_user_id,
+                           created_by_project_id=r.created_by_project_id,
+                           archive_policy_name=ap_name,
+                           name=name,
+                           resource_id=r.id)
+                session.add(m)
+                try:
+                    session.flush()
+                except exception.DBDuplicateEntry:
+                    raise indexer.NamedMetricAlreadyExists(name)
+                except exception.DBReferenceError as e:
+                    if (e.constraint ==
+                       'fk_metric_archive_policy_name_archive_policy_name'):
+                        raise indexer.NoSuchArchivePolicy(ap_name)
+                    raise
+
         session.expire(r, ['metrics'])
 
-    def delete_resource(self, resource_id, delete_metrics=None):
+    def delete_resource(self, resource_id):
         session = self.engine_facade.get_session()
         with session.begin():
-            q = session.query(Resource).filter(
-                Resource.id == resource_id).options(
-                    sqlalchemy.orm.joinedload('metrics'))
-            r = q.first()
-            if r is None:
+            # We are going to delete the resource; the on delete will set the
+            # resource_id of the attached metrics to NULL, we just have to mark
+            # their status as 'delete'
+            session.query(Metric).filter(
+                Metric.resource_id == resource_id).update(
+                    {"status": "delete"})
+            if session.query(Resource).filter(
+                    Resource.id == resource_id).options(
+                        sqlalchemy.orm.joinedload('metrics')).delete() == 0:
                 raise indexer.NoSuchResource(resource_id)
-            if delete_metrics is not None:
-                delete_metrics(self.get_metrics([m.id for m in r.metrics]))
-            q.delete()
 
     def get_resource(self, resource_type, resource_id, with_metrics=False):
         resource_cls = self._resource_type_to_class(resource_type)
@@ -414,7 +443,9 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
             properties={
                 'metrics': sqlalchemy.orm.relationship(
                     Metric,
-                    primaryjoin=Metric.resource_id == stmt.c.id,
+                    primaryjoin=sqlalchemy.and_(
+                        Metric.resource_id == stmt.c.id,
+                        Metric.status == 'active'),
                     foreign_keys=Metric.resource_id)
             })
 
@@ -510,10 +541,16 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
         session.expunge_all()
         return all_resources
 
+    def expunge_metric(self, id):
+        session = self.engine_facade.get_session()
+        if session.query(Metric).filter(Metric.id == id).delete() == 0:
+            raise indexer.NoSuchMetric(id)
+
     def delete_metric(self, id):
         session = self.engine_facade.get_session()
-        session.query(Metric).filter(Metric.id == id).delete()
-        session.flush()
+        if session.query(Metric).filter(
+                Metric.id == id).update({"status": "delete"}) == 0:
+            raise indexer.NoSuchMetric(id)
 
 
 class QueryTransformer(object):
@@ -594,9 +631,20 @@ class QueryTransformer(object):
                 raise indexer.QueryAttributeError(table, field_name)
 
             # Convert value to the right type
-            if value is not None and isinstance(attr.type,
-                                                base.PreciseTimestamp):
-                value = utils.to_timestamp(value)
+            if value is not None:
+                converter = None
+
+                if isinstance(attr.type, base.PreciseTimestamp):
+                    converter = utils.to_timestamp
+                elif (isinstance(attr.type, sqlalchemy_utils.UUIDType)
+                      and not isinstance(value, uuid.UUID)):
+                    converter = utils.ResourceUUID
+
+                if converter:
+                    try:
+                        value = converter(value)
+                    except Exception:
+                        raise indexer.QueryValueError(value, field_name)
 
         return op(attr, value)
 
